import React, {
  ReactNode,
  useMemo,
  useEffect,
  useState,
  useRef,
  Fragment
} from 'react';
import {
  useTable,
  useFlexLayout,
  useSortBy,
  useExpanded,
  useRowSelect,
  usePagination,
  useGlobalFilter,
  Row as RowType
} from 'react-table';
import { Column as ColumnType, TableProps } from './types';
import { variables as theme } from '@cloudflare/style-const';
import {
  CSSWithSelectors,
  applyStaticStyles
} from '@cloudflare/style-container';
import { Trans } from '@cloudflare/intl-react';
import { Div, Span } from '@cloudflare/elements';
import { Link } from '@cloudflare/component-link';

import { Icon } from '@cloudflare/component-icon';
import { Checkbox } from '@cloudflare/component-checkbox';
import { Pagination } from '@cloudflare/component-pagination';
import {
  SortableContainer,
  SortableElement,
  arrayMove
} from 'react-sortable-hoc';
import DragHandle from './DragHandle';
import MoveRowButtons from './MoveRowButtons';

import {
  List,
  AutoSizer,
  CellMeasurer,
  CellMeasurerCache
} from 'react-virtualized';
import useMeasure from 'react-use-measure';
import { ResizeObserver as ResizeObserverPolyfill } from '@juggle/resize-observer';

const ResizeObserver = window.ResizeObserver || ResizeObserverPolyfill;

import { DefaultCell } from './CellStyles';
import HeaderRow from './HeaderRow';
import Row from './Row';

const SortableList = SortableContainer(List);
const SortableRow = SortableElement(Row);

const borderColor = theme.colors.gray[7];

export const EXPANDABLE_ID = 'expandable-id';
export const SELECTABLE_ID = 'selectable-id';
export const DRAGGABLE_ID = 'draggable-id';

const defaultScrollHeight = 700;

let count = 0;

const ScrollWrapper = applyStaticStyles(
  `
  .ReactVirtualized__List {
    outline: none
  }
  .cf-table-row-drag {
    pointer-events: auto !important;
    z-index: 2000 !important;
    box-shadow: 0 0 10px rgba(0, 0, 0, 0.2) !important;
    background-color: ${theme.colors.background} !important;
  }
  .cf-table-row-drag * {
    cursor: grabbing !important;
  }
  `,
  Div
);

export const isRowActions = (columnId: string) =>
  [EXPANDABLE_ID, SELECTABLE_ID, DRAGGABLE_ID].includes(columnId);

const draggableColumn = {
  id: DRAGGABLE_ID,
  width: 50,
  Cell: ({ row, rows, column, columns, allColumns }) => {
    const rowId = `table_${column.getTableId()}_row${
      (row.original as any)?.id || row.getRowProps()?.key
    }`;

    const indexOffset = allColumns.length - columns.length;

    return (
      <DefaultCell px={2} justifyContent="center">
        {column.isDragHandleVisible?.() && (
          <>
            <DragHandle />
            <MoveRowButtons
              handleMoveUpClick={
                row.index > 0
                  ? () =>
                      column.moveRow(rows, {
                        oldIndex: row.index,
                        newIndex: row.index - 1
                      })
                  : undefined
              }
              handleMoveDownClick={
                row.index < rows.length - 1
                  ? () => {
                      column.moveRow(rows, {
                        oldIndex: row.index,
                        newIndex: row.index + 1
                      });
                    }
                  : undefined
              }
              describedBy={columns.map(
                (_col, index) => `${rowId}_cell_${indexOffset + index}`
              )}
            />
          </>
        )}
      </DefaultCell>
    );
  }
};

// @ts-ignore
draggableColumn.Cell.displayName = 'DraggableColumnCell';

const expandableColumn = {
  id: EXPANDABLE_ID,
  width: 26,
  Cell: ({ row }: { row: RowType }) => {
    return (
      <DefaultCell
        px={1}
        style={{
          ...(row.isExpanded ? { transform: 'rotate(90deg)' } : {}),
          transition: 'transform 0.18s ease-out'
        }}
      >
        <Icon type="right" />
      </DefaultCell>
    );
  }
};

// @ts-ignore
expandableColumn.Cell.displayName = 'ExpandableColumnCell';

type SelectableColumnHeaderProps = {
  rows: RowType[];
  pageRows?: RowType[];
  column: ColumnType;
  isRowSelectable?: (row: any) => boolean;
  selectAllKey: string;
};

type SelectableColumnCellProps = {
  row: RowType;
  column: ColumnType;
  isRowSelectable?: (row: any) => boolean;
};

const selectableColumn = {
  id: SELECTABLE_ID,
  Header: ({
    rows,
    column,
    pageRows,
    isRowSelectable,
    selectAllKey
  }: SelectableColumnHeaderProps) => {
    // If table is paginated, page contains only the visible rows
    const selectableRows = (pageRows || rows).filter(
      row => !isRowSelectable || isRowSelectable(row)
    );
    const selectedRows = selectableRows.filter(row => row.isSelected);
    const [isChecked, setIsChecked] = useState(selectedRows.length > 0);
    const isIndeterminate =
      selectedRows.length > 0 && selectedRows.length < selectableRows.length;

    useEffect(() => {
      setIsChecked(selectedRows.length > 0);
    }, [selectedRows]);

    const handleToggleAll = () => {
      const selectedValue = !isChecked;
      setIsChecked(selectedValue);
      return selectableRows.forEach(row =>
        row.toggleRowSelected(selectedValue)
      );
    };

    return (
      <Div display="flex" alignItems="flex-start" mt="2px">
        <Checkbox
          name="toggleAll"
          onChange={handleToggleAll}
          checked={isChecked}
          indeterminate={isIndeterminate}
          disabled={
            !selectableRows.length || column.getCellData?.().disableRowSelect
          }
          value="true"
          label=""
        />
        {!!selectAllKey && (
          <Span ml={3}>
            <Trans id={selectAllKey} />
          </Span>
        )}
      </Div>
    );
  },
  Cell: ({ row, column, isRowSelectable }: SelectableColumnCellProps) => {
    const isDisabled =
      (isRowSelectable && !isRowSelectable(row)) ||
      column.getCellData?.().disableRowSelect;
    return (
      <DefaultCell alignItems="flex-start" pr={0} pt={2}>
        <Checkbox
          name="toggleRow"
          onClick={e => e.stopPropagation()}
          {...row.getToggleRowSelectedProps()}
          {...(isDisabled ? { checked: false } : {})} // Force checked = false if disabled
          disabled={isDisabled}
          value="true"
          label=""
        />
      </DefaultCell>
    );
  }
};

// @ts-ignore
selectableColumn.Header.displayName = 'SelectableColumnHeader';

// @ts-ignore
selectableColumn.Cell.displayName = 'SelectableColumnCell';

const defaultColumn = {
  width: 50,
  Cell: ({ cell: { value } }: { cell: { value: ReactNode } }) => (
    <DefaultCell>{value}</DefaultCell>
  )
};

// @ts-ignore
defaultColumn.Cell.displayName = 'DefaultStyleColumnCell';

const defaultComponent = () => <Fragment />;
const defaultNoop = () => {};
const defaultWrapperStyle: CSSWithSelectors = {
  position: 'relative',
  borderBottom: `1px solid ${borderColor}`
};

const Table = <
  RowData extends object = any,
  Filter = never,
  GlobalFilterProps extends object = never
>({
  rowStyle,
  data,
  columns,
  disableHeaders,
  isLoading,
  hasError,
  isEmpty,
  initialSortBy = [],
  initialSelectedRowIds = {},
  renderEmpty: Empty = defaultComponent,
  renderError: Error = defaultComponent,
  numberOfLoadingRows = 10,
  expandable,
  onlyOneExpanding,
  isRowExpandable,
  isRowSelectable,
  onSortBy = defaultNoop,
  onRowClick = defaultNoop,
  onRowSelect,
  className,
  pageSize,
  GlobalFilter,
  globalFilterProps,
  filterFunction,
  bordered,
  onRowMove,
  disableRowSelect,
  showSelectAllInHeader,
  selectedCountKey,
  disableSelectAll = false,
  autoResetExpanded,
  autoResetPage,
  selectAllKey = 'common.table.select_all',
  clearAllKey = 'common.table.clear_all',
  allSelectedOnPageKey = 'common.table.all_selected_on_page',
  allSelectedKey = 'common.table.all_selected',
  wrapperStyle = defaultWrapperStyle,
  getRowId,
  scrollHorizontally,
  ...wrapperProps
}: TableProps<RowData, Filter, GlobalFilterProps>) => {
  const listRef = useRef();
  const canMoveRows = !!(onRowMove && data && data?.length > 0);

  const [tableId] = useState(() => ++count);

  const dragState = useMemo(
    () => ({
      canDrag: true
    }),
    []
  );

  const [cellData] = useState({ disableRowSelect });
  cellData.disableRowSelect = disableRowSelect;

  const memoizedData = useMemo<RowData[]>(
    () =>
      (isLoading
        ? Array.from({ length: numberOfLoadingRows }, () => ({}))
        : [...(data ?? [])]) as RowData[],
    [data, isLoading, numberOfLoadingRows]
  );

  const [reorderableData, setReorderableData] = useState(memoizedData);

  useEffect(() => setReorderableData(memoizedData), [memoizedData]);

  const memoizedColumns = useMemo<ColumnType<RowData>[]>(
    () =>
      columns.map((column: ColumnType<RowData>) => ({
        ...column,
        disableSortBy: !column.enableSortBy
      })),
    [columns]
  );

  if ([GlobalFilter, filterFunction].filter(item => item).length === 1) {
    console.error(
      'Both GlobalFilter and filterFunction must be specified together'
    );
  }

  const cellMeasurerCache = useMemo(
    () =>
      new CellMeasurerCache({
        // TODO prop
        minHeight: 37,
        fixedWidth: true
      }),
    []
  );

  // Use the cellMeasurer cache to determine the size of table when using
  // virtual scrolling.
  let scrollHeight = 0;
  if (onRowMove && pageSize) {
    for (let ii = 0; ii < pageSize; ++ii) {
      // @ts-ignore
      scrollHeight += cellMeasurerCache._rowHeightCache?.[`${ii}-0`] || 0;
    }
  }
  scrollHeight = scrollHeight || defaultScrollHeight;

  const {
    getTableProps,
    getTableBodyProps,
    gotoPage,
    headerGroups,
    rows,
    page,
    prepareRow,
    // @ts-ignore
    state: { sortBy, selectedRowIds, pageIndex, globalFilter },
    canPreviousPage,
    canNextPage,
    previousPage,
    nextPage,
    preGlobalFilteredRows,
    setGlobalFilter,
    toggleAllRowsSelected,
    toggleAllRowsExpanded
  } = useTable<RowData>(
    {
      data: canMoveRows ? reorderableData : memoizedData,
      columns: memoizedColumns,
      defaultColumn,
      manualSortBy: true,
      disableMultiSort: true,
      autoResetExpanded: !!autoResetExpanded,
      autoResetPage: !!autoResetPage,
      initialState: {
        sortBy: initialSortBy,
        selectedRowIds: initialSelectedRowIds,
        pageSize
      },
      globalFilter: filterFunction,
      // @ts-ignore
      getRowId
    },
    useFlexLayout,
    ...(GlobalFilter ? [useGlobalFilter] : []),
    useSortBy,
    useExpanded,
    useRowSelect,
    ...(pageSize ? [usePagination] : []),
    hooks => {
      if (onRowSelect) {
        hooks.visibleColumns.push(columns => [
          {
            ...selectableColumn,
            width: showSelectAllInHeader ? 120 : 34,
            isRowSelectable,
            getCellData: () => cellData
          },
          ...columns
        ]);
      }

      if (canMoveRows) {
        hooks.visibleColumns.push(columns => [
          {
            ...draggableColumn,
            isDragHandleVisible: () => dragState.canDrag,
            getTableId: () => tableId,
            moveRow: (
              rowsToUpdate,
              {
                newIndex,
                oldIndex
              }: {
                rowsToUpdate: RowType<RowData>[];
                newIndex: number;
                oldIndex: number;
              }
            ) => {
              const updatedData = arrayMove(
                rowsToUpdate.map(row => row.original),
                oldIndex,
                newIndex
              );
              setReorderableData(updatedData);
              onRowMove?.(updatedData, { newIndex, oldIndex });
            }
          },
          ...columns
        ]);
      }

      if (expandable) {
        hooks.visibleColumns.push(columns => [expandableColumn, ...columns]);
      }
    }
  );

  useEffect(() => {
    if (!onRowSelect || isLoading) return;

    const rowObjects = Object.keys(selectedRowIds).reduce<RowData[]>(
      (acc, rId) => {
        const rawRow = (
          (preGlobalFilteredRows as RowType<RowData>[]) || rows
        ).find(({ id }) => rId === id);

        if (!rawRow) return acc;
        if (isRowSelectable && !isRowSelectable(rawRow)) return acc;

        return [...acc, { rowIndex: rawRow.index, ...rawRow.original }];
      },
      []
    );

    onRowSelect(rowObjects);
  }, [selectedRowIds, isLoading]);

  useEffect(() => {
    onSortBy(sortBy);
  }, [sortBy]);

  useEffect(() => {
    cellMeasurerCache?.clearAll();
  }, [columns, data]);

  const [ref, { width: containerWidth }] = useMeasure({
    polyfill: ResizeObserver
  });

  const selectedRowsCount = Object.keys(selectedRowIds).length;
  const allSelected = selectedRowsCount === rows.length;
  const pageSelected = !!pageSize && page.every(row => row.isSelected);
  const filtered =
    !!preGlobalFilteredRows && preGlobalFilteredRows.length !== rows.length;
  const showSelectAll = !!selectedRowsCount && pageSelected;

  return (
    <Div {...wrapperProps}>
      {!!GlobalFilter && (
        <GlobalFilter
          rows={rows}
          preGlobalFilteredRows={preGlobalFilteredRows}
          globalFilter={globalFilter}
          selectedRowsCount={selectedRowsCount}
          toggleAllRowsSelected={toggleAllRowsSelected}
          setGlobalFilter={filter => {
            dragState.canDrag = !filter;
            if (!!pageSize) {
              gotoPage(0);
            }
            setGlobalFilter(filter);
          }}
          {...(globalFilterProps || {})}
        />
      )}

      <div
        ref={ref}
        className={className}
        style={{
          ...(bordered ? { border: `1px solid ${borderColor}` } : {}),
          ...wrapperStyle,
          ...(scrollHorizontally ? { overflowX: 'auto' } : {})
        }}
      >
        <Div {...getTableProps()}>
          <HeaderRow
            disableHeaders={disableHeaders}
            headerGroups={headerGroups}
            isLoading={isLoading}
            isEmpty={isEmpty}
            hasError={hasError}
            selectAllKey={showSelectAllInHeader ? selectAllKey : undefined}
          />

          {(showSelectAll || selectedCountKey) &&
            !disableRowSelect &&
            !disableSelectAll && (
              <Div
                backgroundColor="gray.9"
                textAlign="center"
                padding={2}
                borderRadius={2}
                borderBottom="1px solid"
                borderColor={borderColor}
              >
                <Trans
                  id={
                    selectedCountKey ||
                    (allSelected ? allSelectedKey : allSelectedOnPageKey)
                  }
                  smartCount={
                    !selectedCountKey && pageSize && !allSelected
                      ? page.filter(row => row.isSelected).length
                      : selectedRowsCount
                  }
                />
                <Link
                  pl={3}
                  onClick={e => {
                    e.preventDefault();
                    toggleAllRowsSelected(!allSelected);
                  }}
                >
                  <Trans id={allSelected ? clearAllKey : selectAllKey} />
                </Link>
              </Div>
            )}

          <Div {...getTableBodyProps()}>
            {((!isLoading && isEmpty) || (filtered && !rows.length)) && (
              <Empty globalFilter={globalFilter} />
            )}
            {!isLoading && hasError && <Error />}
            {(!isEmpty || !hasError) &&
              (!filtered && canMoveRows ? (
                <ScrollWrapper style={{ height: scrollHeight + 'px' }}>
                  <AutoSizer>
                    {({ width }: { width: number }) => (
                      <SortableList
                        ref={listRef}
                        deferredMeasurementCache={cellMeasurerCache}
                        rowHeight={cellMeasurerCache.rowHeight}
                        height={scrollHeight}
                        rowCount={rows.length}
                        width={width}
                        rowRenderer={({
                          index,
                          style,
                          parent
                        }: {
                          index: number;
                          style: object;
                          parent: any;
                        }) => {
                          const row = rows[index];

                          prepareRow(row);

                          // Use row.original.id as key instead of row.getRowProps()?.key to fix the bug causing by index shiftting
                          // ex: row.original.id => 911874bcebe89ca25c0d9e415de7f634, row.getRowProps()?.key => row_0
                          const rowId = `table_${tableId}_row${
                            (row.original as RowType<RowData>).id ||
                            row.getRowProps()?.key
                          }`;
                          return (
                            <CellMeasurer
                              cache={cellMeasurerCache}
                              columnIndex={0}
                              key={`sortable-row-${rowId}`}
                              parent={parent}
                              rowIndex={index}
                            >
                              <SortableRow
                                row={row}
                                rowId={rowId}
                                index={index}
                                rowIndex={index}
                                style={style}
                                rowStyle={{ ...rowStyle, ...style }}
                                isLoading={isLoading}
                                striped={false}
                                rowHighlight={false}
                                onRowClick={onRowClick}
                                containerWidth={containerWidth}
                                disabled={
                                  (isRowSelectable && !isRowSelectable(row)) ||
                                  disableRowSelect
                                }
                              />
                            </CellMeasurer>
                          );
                        }}
                        lockAxis="y"
                        helperClass="cf-table-row-drag"
                        useDragHandle={true}
                        onSortEnd={({
                          newIndex,
                          oldIndex
                        }: {
                          newIndex: number;
                          oldIndex: number;
                        }) => {
                          cellMeasurerCache.clearAll();
                          const updatedData = arrayMove(
                            reorderableData,
                            oldIndex,
                            newIndex
                          );
                          setReorderableData(updatedData);
                          onRowMove?.(updatedData, { newIndex, oldIndex });
                        }}
                      />
                    )}
                  </AutoSizer>
                </ScrollWrapper>
              ) : (
                (page || rows).map((row, i) => {
                  prepareRow(row);
                  // Use row.original.id as key instead of row.getRowProps()?.key to fix the bug causing by index shiftting
                  // ex: row.original.id => 911874bcebe89ca25c0d9e415de7f634, row.getRowProps()?.key => row_0
                  const rowId = `table_${tableId}_row${
                    (row.original as RowType<RowData>).id ||
                    row.getRowProps()?.key
                  }`;
                  return (
                    <Row
                      key={rowId}
                      testId={`row_${(row.original as RowType<RowData>).id}`}
                      row={row}
                      rowIndex={i}
                      rowId={rowId}
                      rowStyle={rowStyle}
                      expandable={expandable}
                      isRowExpandable={isRowExpandable}
                      isRowSelectable={isRowSelectable}
                      onlyOneExpanding={onlyOneExpanding}
                      toggleAllRowsExpanded={toggleAllRowsExpanded}
                      isLoading={isLoading}
                      striped={false}
                      rowHighlight={false}
                      onRowClick={onRowClick}
                      containerWidth={containerWidth}
                      disabled={isRowSelectable && !isRowSelectable(row)}
                    />
                  );
                })
              ))}
          </Div>
        </Div>
        {!onRowMove && !!pageSize && rows.length > pageSize && (
          <Div
            borderTop={bordered ? 'none' : '1px solid'}
            borderColor={borderColor}
          >
            <Pagination
              p={2}
              loading={isLoading}
              start={pageIndex * pageSize + 1}
              end={Math.min(pageIndex * pageSize + pageSize, rows.length)}
              total={rows.length}
              onPrevClick={previousPage}
              onNextClick={nextPage}
              disablePreviousButton={!canPreviousPage}
              disableNextButton={!canNextPage}
            />
          </Div>
        )}
      </div>
    </Div>
  );
};

export default Table;
